/*======================================================================*/
/*  File:  Virtual.cpp                                                  */
/*  Virtual - Demonstrate VIRTUAL functions including use of            */
/*            VIRTUAL destructors and dynamic calls                     */
/*======================================================================*/

#include <iostream.h>

/*======================================================================*/

class Shape {
  public:
    Shape ()                   { cout << "CTOR - Shape" << endl; }
    virtual ~Shape ()          { cout << "DTOR - Shape" << endl; }
    void Doit (Shape *sp)      { cout << ((sp) ? sp->What() : " NULL") << endl; }
    void Dtor (Shape *sp)      { sp->~Shape(); }
    virtual char *What (void)  { return " SHAPE"; };
};

/*======================================================================*/

class Box : public Shape {
  public:
    Box()              { cout << "CTOR - Box" << endl; }
    ~Box ()            { cout << "DTOR - Box" << endl; }
    char *What (void)  { return " BOX"; }
};

/*======================================================================*/

class Circle : public Shape {
  public:
    Circle ()          { cout << "CTOR - Circle" << endl; }
    ~Circle ()         { cout << "DTOR - Circle" << endl; }
    char *What (void)  { return " CIRCLE"; }
};

/*======================================================================*/

int main (void) {
    cout << "Demonstrate VIRTUAL functions and destructors" << endl << endl;

    {
    cout << "Initial constructor calls" << endl;
    Shape   s;
    Box     b;
    Circle  c;
    Shape   *sp[] = { &s, &b, &c };


    cout << endl << "Demonstrate VIRTUAL functions" << endl << endl;

    c.Doit (&b);  
    c.Doit (&c);  
    cout << endl;
    
    b.Doit (&b);  
    b.Doit (&c);  
    cout << endl;

    for (int i = 0; i < sizeof sp / sizeof sp[0]; i++)
        cout << sp[i]->What() << endl;
    cout << endl;


    cout << "Demonstrate VIRTUAL DESTRUCTOR calls" << endl;

    c.Dtor (new Shape ());   cout << endl;
    b.Dtor (new Box ());     cout << endl;
    s.Dtor (new Circle ());  cout << endl;

    cout << "Final destructor calls" << endl;
    }

    return 0;
}

/*======================================================================*/

/* .....  Output from the above program...

Demonstrate VIRTUAL functions and destructors

Initial constructor calls
CTOR - Shape
CTOR - Shape
CTOR - Box
CTOR - Shape
CTOR - Circle

Demonstrate VIRTUAL functions

 BOX
 CIRCLE

 BOX
 CIRCLE

 SHAPE
 BOX
 CIRCLE

Demonstrate VIRTUAL DESTRUCTOR calls
CTOR - Shape
DTOR - Shape

CTOR - Shape
CTOR - Box
DTOR - Box
DTOR - Shape

CTOR - Shape
CTOR - Circle
DTOR - Circle
DTOR - Shape

Final destructor calls
DTOR - Circle
DTOR - Shape
DTOR - Box
DTOR - Shape
DTOR - Shape

..... end of output ... */

